#ifndef PORYGONLANG_MATHLIBRARY_HPP
#define PORYGONLANG_MATHLIBRARY_HPP

#include <utility>
#include <cmath>
#include <cfloat>
#include "../Evaluator/EvalValues/EvalValue.hpp"
#include "../Evaluator/EvalValues/NumericEvalValue.hpp"

#include "../UserData/UserDataScriptType.hpp"
#include "../ScriptTypes/FunctionScriptType.hpp"
#include "../ScriptTypes/ScriptType.hpp"
#include "../UserData/UserDataFunctionType.hpp"
#include "../Evaluator/EvalValues/ScriptFunctionEvalValue.hpp"
#include "../UserData/UserDataFunction.hpp"
#include "../Utilities/Random.hpp"

#ifndef WINDOWS
__asm__(".symver log,log@GLIBC_2.2.5");
__asm__(".symver exp,exp@GLIBC_2.2.5");
#endif

namespace Porygon::StandardLibraries {
    using namespace Porygon::Evaluation;
    using namespace Porygon::Utilities;
    class MathLibrary {
        typedef UserData::UserDataReturnValue Return;

        //region templates

#define FLOAT_TYPE NumericScriptType::AwareFloat
#define INTEGER_TYPE NumericScriptType::AwareInt
#define BOOL_TYPE ScriptType::BoolType
#define FUNCTION(fieldName) \
        [](void* obj) -> const EvalValue* { \
                auto t = new Porygon::Evaluation::GenericFunctionEvalValue(make_shared<GenericFunctionScriptType>(), \
                        Porygon::Utilities::Random::Get()); \
                t->RegisterOption(new Porygon::UserData::UserDataFunction(fieldName, nullptr)); \
                return t;}, \
        nullptr
        //endregion

        inline static const Return* RETURN(double d){
            return new Return(new NumericEvalValue(d));
        }
        inline static const Return* RETURN(int64_t i){
            return new Return(new NumericEvalValue(i));
        }


        static const Return* _abs(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount){
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            if (parameter->IsFloat()){
                return RETURN(std::abs(parameter->EvaluateFloat()));
            } else{
                return RETURN(std::abs(parameter->EvaluateInteger()));
            }
        }

        static const Return* _acos(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            return RETURN(std::acos(parameter->EvaluateFloat()));
        }

        static const Return* _asin(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            return RETURN(std::asin(parameter->EvaluateFloat()));
        }

        static const Return* _atan(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            if (parameterCount == 1){
                auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
                return RETURN(std::atan(parameter->EvaluateFloat()));
            }
            else{
                auto parameter1 = dynamic_cast<const NumericEvalValue*>(parameters[0]);
                auto parameter2 = dynamic_cast<const NumericEvalValue*>(parameters[1]);
                return RETURN(std::atan2(parameter1->EvaluateFloat(), parameter2->EvaluateFloat()));
            }
        }
        static const Return* _ceil(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            auto f = parameter->EvaluateFloat();
            return RETURN(static_cast<int64_t>(std::ceil(f)));
        }

        static const Return* _cos(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            auto f = parameter->EvaluateFloat();
            return RETURN(std::cos(f));
        }

        static const Return* _deg(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            auto f = parameter->EvaluateFloat();
            return RETURN(f * 180 / M_PI);
        }

        static const Return* _exp(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            auto f = parameter->EvaluateFloat();
            return RETURN(std::exp(f));
        }

        static const Return* _floor(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            auto f = parameter->EvaluateFloat();
            return RETURN(std::floor(f));
        }

        static const Return* _fmod(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto f1 = dynamic_cast<const NumericEvalValue*>(parameters[0])->EvaluateFloat();
            auto f2 = dynamic_cast<const NumericEvalValue*>(parameters[0])->EvaluateFloat();
            return RETURN(std::fmod(f1, f2));
        }

        static const EvalValue* _huge;

        static const Return* _log(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount){
            if (parameterCount == 1){
                auto f = dynamic_cast<const NumericEvalValue*>(parameters[0])->EvaluateFloat();
                return RETURN(std::log(f));
            }
            else{
                auto f1 = dynamic_cast<const NumericEvalValue*>(parameters[0])->EvaluateFloat();
                auto f2 = dynamic_cast<const NumericEvalValue*>(parameters[1])->EvaluateFloat();
                return RETURN(std::log(f2) / std::log(f1));
            }
        }

        static const EvalValue* _maxInteger;
        static const EvalValue* _minInteger;
        static const EvalValue* _pi;

        static const Return* _rad(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            auto f = parameter->EvaluateFloat();
            return RETURN(M_PI *(f / 180));
        }

        static const Return* _sin(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            return RETURN(std::sin(parameter->EvaluateFloat()));
        }

        static const Return* _sqrt(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            return RETURN(std::sqrt(parameter->EvaluateFloat()));
        }

        static const Return* _tan(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter = dynamic_cast<const NumericEvalValue*>(parameters[0]);
            return RETURN(std::tan(parameter->EvaluateFloat()));
        }

        static const Return* _ult(void*, const ScriptOptions*, const Evaluation::EvalValue* parameters[], int parameterCount) {
            auto parameter1 = dynamic_cast<const NumericEvalValue*>(parameters[0])->EvaluateInteger();
            auto parameter2 = dynamic_cast<const NumericEvalValue*>(parameters[1])->EvaluateInteger();
            return new Return(new BooleanEvalValue(parameter1 < parameter2));
        }

        static UserData::UserDataFunctionOption* CreateFunctionOption(const shared_ptr<const ScriptType>& returnType,
                const std::vector<shared_ptr<const ScriptType>>& params){
            return new UserData::UserDataFunctionOption(returnType, params);
        }


    public:
        static UserData::UserData* CreateUserData(){
            return new UserData::UserData(new Utilities::HashedString(new u16string(u"__math__")), {
                {
                    HashedString::ConstHash("abs"),
                 new UserData::UserDataField(
                         (new GenericFunctionScriptType())
                             ->RegisterFunctionOption(0, CreateFunctionOption(INTEGER_TYPE, {INTEGER_TYPE}))
                             ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                         , FUNCTION(_abs))
                },
                {
                    HashedString::ConstHash("acos"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_acos))
                },
                {
                    HashedString::ConstHash("asin"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_asin))
                },
                {
                    HashedString::ConstHash("atan"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE, FLOAT_TYPE}))
                            , FUNCTION(_atan))
                },
                {
                    HashedString::ConstHash("ceil"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(INTEGER_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_ceil))
                },
                {
                    HashedString::ConstHash("cos"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_cos))
                },
                {
                    HashedString::ConstHash("deg"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_deg))
                },
                {
                    HashedString::ConstHash("exp"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_exp))
                },
                {
                    HashedString::ConstHash("floor"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(INTEGER_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_floor))
                },
                {
                    HashedString::ConstHash("fmod"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE, FLOAT_TYPE}))
                            , FUNCTION(_fmod))
                },
                {
                    HashedString::ConstHash("huge"),
                    new UserData::UserDataField(
                            FLOAT_TYPE
                            ,  [](void* obj) -> const Porygon::Evaluation::EvalValue* {return _huge;}, nullptr)
                },
                {
                    HashedString::ConstHash("log"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE, FLOAT_TYPE}))
                            , FUNCTION(_log))
                },
                {
                    HashedString::ConstHash("maxinteger"),
                    new UserData::UserDataField(
                            INTEGER_TYPE
                            ,  [](void* obj) -> const Porygon::Evaluation::EvalValue* {return _maxInteger;}, nullptr)
                },
                {
                    HashedString::ConstHash("mininteger"),
                    new UserData::UserDataField(
                            INTEGER_TYPE
                            ,  [](void* obj) -> const Porygon::Evaluation::EvalValue* {return _minInteger;}, nullptr)
                },
                {
                    HashedString::ConstHash("pi"),
                    new UserData::UserDataField(
                            FLOAT_TYPE
                            ,  [](void* obj) -> const Porygon::Evaluation::EvalValue* {return _pi;}, nullptr)
                },
                {
                    HashedString::ConstHash("rad"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_rad))
                },
                {
                    HashedString::ConstHash("sin"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_sin))
                },
                {
                    HashedString::ConstHash("sqrt"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_sqrt))
                },
                {
                    HashedString::ConstHash("tan"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(FLOAT_TYPE, {FLOAT_TYPE}))
                            , FUNCTION(_tan))
                },
                {
                    HashedString::ConstHash("ult"),
                    new UserData::UserDataField(
                            (new GenericFunctionScriptType())
                                    ->RegisterFunctionOption(0, CreateFunctionOption(BOOL_TYPE, {INTEGER_TYPE, INTEGER_TYPE}))
                            , FUNCTION(_ult))
                },

          });
        }
    };
}

#endif //PORYGONLANG_MATHLIBRARY_HPP
